#!/bin/sh

set -e

. ./cleanup-handler.sh

# debugging:
# remove 'check' in front of failing command
# 'set -x' changes output and 'check' might fail
# -> disable output checking: expect empty string ""

UNITTEST=1
. ./sdo-disable-logs.sh # runs with dash

ROOTFS=/tmp/rootfs
STATUS=/tmp/sdo-disabled-logs.status

CRITICAL_USAGE_PERCENT=70 # critical if usage higher than
CRITICAL_FREE_SPACE=5    # critical if less space available

# rm $STATUS

target()
{
    grep -q ARM926EJ /proc/cpuinfo
}

if target; then
    echo "Running on target"
    # target has no sudo
    sudo () {
        $*
    }

    # /tmp is a ramdisk, redirect to disk if RAM is scarce
    mkdir -p /alt_tmp
    export TMPDIR=/alt_tmp # non-ramdisk

    if [ $(opkg list-installed 'kernel-module-loop' | wc -l) -eq 0 ]; then
        echo "need kernel-module-loop to mount virtual sdcard, install? yes/no"
        while read a; do
            case $a in
                "yes")
                    opkg install kernel-module-loop
                    break;;
                "no")
                    echo "... quitting"
                    exit;;
            esac
            echo "need kernel-module-loop to mount virtual sdcard, install? yes/no"
        done
    fi
    # not automatically done by mount command
    modprobe loop || die "failed to load kernel-module-loop"
else
    echo "Running on host"
fi

# for debugging:
debug_wait()
{
    while read a; do
        [ -n "$a" -a "$a" = 'quit' ] && break
    done
}

test_passed()
{
    echo "PASSED"
}

check()
{
    fct=$1
    exp_ret=$2
    exp_output=$3
    ret=0
    # eval runs in a separate shell, it's environment
    # is not visible from this shell
    output=`eval $fct 2>&1` || ret=$?
    [ $ret -eq $exp_ret ] || die "$fct: return code mismatch [$exp_ret != $ret]"

    if [ -z "$exp_output" ]; then
        return
    fi

    [ "$output" = "$exp_output" ] || \
        die "$fct: output mismatch [\'$exp_output\' != \'$output\']"
}
echo "######################################"
echo "TEST: cleanup-handler"
echo "######################################"
touch /tmp/foo /tmp/bar
cleanup_push "rm /tmp/foo"
cleanup_push "rm /tmp/bar"
cleanup
check 'test -f /tmp/foo' 1 ""
check 'test -f /tmp/bar' 1 ""
test_passed

if target; then
    echo "######################################"
    echo "TEST: runit_service"
    echo "######################################"
    check "runit_service dss" 0 ""
    check "runit_service runit.sh" 1 ""
    test_passed

    echo "######################################"
    echo "TEST: lookup_logdir"
    echo "######################################"
    for D in $MANAGED_DAEMONS; do
        check "lookup_logdir $d" 0 "$ROOTFS/var/log/$d"
        check "test -d /var/log/$d" 0 ""
    done
    check "lookup_logdir sdcard-overlay" 0 "$ROOTFS/var/log/sdcard"
    check "test -d /var/log/sdcard" 0 ""
    test_passed

    echo "######################################"
    echo "TEST: MANAGED_DAEMONS init.d/runit detection"
    echo "######################################"
    for D in $MANAGED_DAEMONS; do
        if runit_service $D; then
            continue
        fi
        if initd_service $D; then
            continue
        fi
        if [ $D = "dss-addons" ]; then
            continue
        fi
        echo "Unknown managed daemon $D"
    done
    test_passed

    echo "######################################"
    echo "TEST: disable svlogd MANAGED_DAEMONS"
    echo "######################################"
    for D in $MANAGED_DAEMONS; do
        if ! is_runit_svlogd $D; then
            continue
        fi
        LOGDIR=`lookup_logdir $D`
        LOGDIR=${LOGDIR#$ROOTFS}
        echo "checking logdir: $LOGDIR"
        runit_svlogd $D stop
        for i in `seq 180`; do
            fuser ${LOGDIR#$ROOTFS} > /dev/null || break
            echo "passed:${i}s open fd: `fuser ${LOGDIR#$ROOTFS}`"
            sleep 1
        done
        check "fuser ${LOGDIR#$ROOTFS}" 1 # -> no more open fd
        echo "disable works"
        runit_svlogd $D start
        for i in `seq 180`; do
            fuser ${LOGDIR#$ROOTFS} > /dev/null && break
            echo "passed:${i}s open fd: `fuser ${LOGDIR#$ROOTFS}`"
            sleep 1
        done
        check "fuser ${LOGDIR#$ROOTFS}" 0 # -> has open fd
        echo "restore works"
    done
    test_passed
fi

loopback_disk()
{
    img=$(mktemp)
    mntpnt=$1
    size=$(($CRITICAL_FREE_SPACE * 2 * 1024))  # MB

    echo "-- create virtual sd-card"
    [ -n "img" ] || die "failed to create temporary file name"
    dd if=/dev/zero of=$img bs=1024 count=$size > /dev/null 2>&1
    cleanup_push "rm $img"

    /sbin/mkfs.ext2 -F $img > /dev/null 2>&1
    [ -d "$mntpnt" ] && die "$mntpnt already exists"
    mkdir $mntpnt
    cleanup_push "rmdir $mntpnt"
    sudo mount -o loop=/dev/loop0 $img $mntpnt
    [ $? -eq 0 ] || die "failed to mount loopback device"
    cleanup_push "sudo umount $mntpnt"

    sudo chown -R `id -u`:`id -g` $mntpnt
}

echo "######################################"
echo "TEST: create virtual sd-card, functions for fs usage in percent and MB"
echo "######################################"

loopback_disk $ROOTFS
check enough_free_space 0 ""
echo "used : `fs_pcent_used`% | free : `fs_free_mb` MB"

# less than CRITICAL_FREE_SPACE
dd if=/dev/zero of=$ROOTFS/foo bs=1024 count=$((1024 * 6))> /dev/null 2>&1
check enough_free_space 2 ""

# exceed CRITICAL_USAGE_PERCENT
dd if=/dev/zero of=$ROOTFS/dig bs=1024 count=$((1024 * 2))> /dev/null 2>&1
check enough_free_space 1 ""

# start-over
test_passed

# manual cleanup, faster, mind lost+found
rm -r $ROOTFS/* || true

echo "######################################"
echo "TEST: mount read-only overlay and verify it's not writable"
echo "######################################"
# inherit loopback from previous test

check "mount_ro_aufs $ROOTFS/foo" 2 "$ROOTFS/foo does not exist"
check "mountpoint -q $ROOTFS/foo" 1 ""
mkdir -p $ROOTFS/foo
check "mount_ro_aufs $ROOTFS/foo" 0 ""
check "mountpoint -q $ROOTFS/foo" 0 ""
check "touch $ROOTFS/foo/bar" 1 ""
test_passed

# cleanup manually
sudo umount $ROOTFS/foo
rm -r $ROOTFS/* || true

echo "######################################"
echo "TEST: exceed usage level, disable log folders to reduce usage"
echo "######################################"
# inherit loopback from previous test

mockup_logs()
{
    logdir=$1
    size=${2:-512} # kBytes
    name=${3:-log}
    mkdir -p $logdir
    dd if=/dev/zero of=$logdir/$name bs=1024 count=$size > /dev/null 2>&1 || true
}

# fill fs till 100%
MANAGED_DAEMONS=""
for i in `seq -w 20 1 60`; do
    MANAGED_DAEMONS="$MANAGED_DAEMONS $i"
    LOGDIR=`lookup_logdir $i`
    mockup_logs $LOGDIR 512 log
    if [ `fs_pcent_used` -eq 100 ]; then
        break
    fi
done

check enough_free_space 1 ""
echo "used : `fs_pcent_used`% | free : `fs_free_mb` MB"

cleanup_push "rm $STATUS" # created by select_victim
umount_aufs() {
    for D in `cat $STATUS`; do
        LOGDIR=`lookup_logdir $D`
        sudo umount $LOGDIR 2> /dev/null || true
    done
}
cleanup_push "umount_aufs"

# stop daemons to make sure no open fd, that would prevent
# the fs from freeing the blocks
select_victim
test_passed

echo "######################################"
echo "TEST: check read-only folders, not writable by root"
echo "######################################"
# inherit loopback from previous test

# verify read-only directory are not writable
old_free=`fs_free_mb`
old_pcent=`fs_pcent_used`
echo "before: $old_pcent% | free : $old_free MB"

# try to write to now read-only folders
for D in `cat $STATUS`; do
    LOGDIR=`lookup_logdir $D`
    mockup_logs $LOGDIR 1
done

echo "after : `fs_pcent_used`% | free : `fs_free_mb` MB"
check fs_free_mb 0 $old_free  # usage didn't change
check fs_pcent_used 0 $old_pcent
test_passed

echo "######################################"
echo "TEST: grow remaining writable logs, then disable more logs to reduce usage"
echo "######################################"
# inherit loopback from previous test

# enlarge still writable logs
for D in $MANAGED_DAEMONS; do
    LOGDIR=`lookup_logdir $D`
    if mountpoint -q $LOGDIR; then
        # already disabled
        continue
    fi
    echo $LOGDIR not yet disabled, grow log files
    for i in `seq -w 0 5`; do
        mockup_logs $LOGDIR 512 addon$i
    done
done

echo "used : `fs_pcent_used`% | free : `fs_free_mb` MB"
check enough_free_space 1 ""

# re-run algorithm
select_victim

echo "used : `fs_pcent_used`% | free : `fs_free_mb` MB"
check enough_free_space 0 ""
test_passed

echo "######################################"
echo "TEST: invalid MANAGED_DAEMONS config "
echo "######################################"

# enlarge still writable logs
for D in $MANAGED_DAEMONS; do
    LOGDIR=`lookup_logdir $D`
    if mountpoint -q $LOGDIR; then
        # already disabled
        continue
    fi
    echo $LOGDIR not yet disabled, grow log files
    for i in `seq -w 0 5`; do
        mockup_logs $LOGDIR 512 addon$i
    done
done

MANAGED_DAEMONS_ORIG=${MANAGED_DAEMONS}
MANAGED_DAEMONS="brumbal ${MANAGED_DAEMONS}"

select_victim
test_passed

echo
echo
echo  "     ============="
echo  "     |  SUCCESS  |"
echo  "     ============="
echo
